package trees;

import compare.Comparable;
import java.util.Vector;
import java.awt.*;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
class OrderedTreeNode
	{
	Comparable key;
	float x, y;
	OrderedTreeNode parent;
	OrderedTreeNode firstChild, lastChild;
	OrderedTreeNode rightSibling, leftSibling;
	OrderedTreeNode leftCousin;
	int numChildren;

	///////////////////////////////////////////////////////////////////////
	OrderedTreeNode(Comparable key)
		{
		/*
		Create a new ordered tree node.
		Each node has an (x,y) position, a key field, and five pointer fields:
		firstChild, lastChild, rightSibling, leftSibling, and parent.
		*/

		this.key = key;
		x = 0; y = 0;
		parent = null;
		firstChild = null; lastChild = null;
		rightSibling = null; leftSibling = null;
		leftCousin = null;
		numChildren = 0;
		}
	}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
public class OrderedTree extends Frame
	{
	/**
	Implement an ordered tree on Comparable keys.
	This implementation assumes the existence of Comparable objects.
	It extends Frame so that it can create its own displayable images.

	Public Nethods:
		OrderedTree()	create a new tree
		makeEmpty()		empty this tree
		isEmpty()		is this tree empty?
		isPresent()		is one particular Comparable array in this tree?
		insert()		insert a given Comparable array into this tree
		delete()		delete a particular Comparable array from this tree
		display()		display this tree using a given graphics context
	*/

	private OrderedTreeNode root = null;//this tree's root
	private int numNodes = 0;			//the number of nodes in this tree
	private Vector walk;				//used when walking the tree

	//////////////////////////////////////////////////////////////////////
	public OrderedTree()
		{
		/*
		Create a new (and empty) ordered tree.
		*/

		root = null;
		numNodes = 0;
		}

	//////////////////////////////////////////////////////////////////////
	public void makeEmpty()
		{
		/*
		Empty this tree.
		*/

		root = null;
		numNodes = 0;
		}

	//////////////////////////////////////////////////////////////////////
	public boolean isEmpty()
		{
		/*
		Tell the world whether or not this tree is empty.
		*/

		if (root == null)
			return true;
		else
			return false;
		}

	//////////////////////////////////////////////////////////////////////
	public boolean isPresent(Comparable[] nodeList)
		{
		/*
		Tell the world whether a node described by nodeList is in this tree.
		*/

		//if the list is empty, no node has been specified, so return false
		if (nodeList.length == 0)
			return false;

		//check the tree level by level
		OrderedTreeNode node = root;
		for (int level = 0; level < nodeList.length; level++)
			{
			//find a node with the current key on this level
			node = checkSiblings(nodeList[level], node);

			//if there's no such node, the search failed, so return false
			if (node == null)
				return false;

			//we've found the current key at this level, so drop to next level
			node = node.firstChild;
			}

		//we've fallen out, so the search succeeded
		return true;
		}

	//////////////////////////////////////////////////////////////////////
	private OrderedTreeNode checkSiblings(Comparable value,
										OrderedTreeNode node)
		{
		/*
		Return whichever sibling of node whose key matches value;
		if there is no such node, return null.
		*/

		while ((node != null) && (value.compareTo(node.key) != 0))
			node = node.rightSibling;

		return node;
		}

	//////////////////////////////////////////////////////////////////////
	public void insert(Comparable[] nodeList)
		{
		/*
		Insert a node described by nodeList into this tree.
		*/

		//if the list is empty, there's nothing to do
		if (nodeList.length == 0)
			return;

		//create the new node to be inserted
		OrderedTreeNode newNode = new
							OrderedTreeNode(nodeList[nodeList.length - 1]);

		//if this tree is empty, make the new node its root and return
		if (root == null)
			{
			root = newNode;
			numNodes++;
			return;
			}

		//find the penultimate node in nodeList
		OrderedTreeNode node = root;
		OrderedTreeNode parentNode = null;
		for (int level = 0; level < nodeList.length - 1; level++)
			{
			node = checkSiblings(nodeList[level], node);

			//if the search failed, one of the new node's ancestors is wrong,
			//silently return failure
			if (node == null)
				return;

			//save the current node, and drop down one more level
			parentNode = node;
			node = node.firstChild;
			}

		//okay, parentNode now points to the soon-to-be parent of the new node
		//link newNode into this tree
		newNode.parent = parentNode;
		parentNode.numChildren++;
		if (parentNode.lastChild != null)
			{
			newNode.leftSibling = parentNode.lastChild;
			parentNode.lastChild.rightSibling = newNode;
			parentNode.lastChild = newNode;
			}
		else
			{
			parentNode.firstChild = newNode;
			parentNode.lastChild = newNode;
			}

		//increase the number of nodes in this tree
		numNodes++;
		}

	//////////////////////////////////////////////////////////////////////
	public void delete(Comparable[] nodeList)
		{
		/*
		Delete the node described by nodeList from this tree.
		*/

		//if either the list or this tree is empty, there's nothing to do
		if ((nodeList.length == 0) || (root == null))
			return;

		//if we're about to delete the root, handle it specially
		if (nodeList.length == 1)
			{
			if (nodeList[0].compareTo(root.key) == 0)
				makeEmpty();
			return;
			}

		//find the node to be deleted
		OrderedTreeNode node = root.firstChild;
		OrderedTreeNode parentNode = root;
		for (int level = 1; level < nodeList.length; level++)
			{
			node = checkSiblings(nodeList[level], node);

			//if the search failed, nodeList is inaccurate,
			//silently return failure
			if (node == null)
				return;

			parentNode = node;
			node = node.firstChild;
			}

		//okay, parentNode now points to the node to be deleted

		//if it's childless, just delete it
		if (parentNode.firstChild == null)
			{
			if ((parentNode.parent).firstChild == parentNode)
				parentNode.parent.firstChild = null;
			}
		else
			{
			//if it has one child, bump up its child one level
			if (parentNode.firstChild == parentNode.lastChild)
				{
				}
			//else it has more than one child
			else
				{
				}
			}

		//decrease the number of nodes in this tree
		numNodes--;
		}

	//////////////////////////////////////////////////////////////////////
	public void display(Graphics g)
		{
		/*
		Display this tree using the graphics context g.
		*/

		if (root == null)
			return;

		//set each node's tentative position
		initializeNodePositions();

		//adjust each node's tentative position
		adjustNodePositions(root, root.x, "");

		//reset the node positions so that the tree is translated left and up
		reset();

		//set the color and font for keys
		g.setColor(Color.red);
		g.setFont(new Font("Times Roman", Font.PLAIN, 3));

		//get the fontmetrics and find the longest key's display dimensions
		FontMetrics fm = g.getFontMetrics();
		int nodeWidth = printTree(root, "    ",
								fm, fm.stringWidth(root.key.toString()));

		int nodeHeight = fm.getHeight();
		int labelOffset = fm.getAscent() + fm.getLeading();

		//recursively draw this tree
		displayTree(g, root, nodeWidth, nodeHeight, labelOffset, fm);
		}

	//////////////////////////////////////////////////////////////////////
	public Dimension treeSize()
		{
		/*
		Return this tree's display dimensions
		*/

		Point XY = findDisplayedSize();
		
		Dimension size = new Dimension(0, 0);
		//resize those dimensions to be more reasonable for display
		size.width = XY.x * 40;
		size.height = XY.y * 40;

		return size;
		}

	//////////////////////////////////////////////////////////////////////
	private int printTree(OrderedTreeNode node, String indentation,
							FontMetrics fm, int nodeWidth)
		{
		/*
		Do a preorder walk of this tree printing each node then its children.
		*/

		if (node != null)
			{
			System.out.println(indentation + node.key + "," + node.x);
			int width = fm.stringWidth(node.key.toString());
			if (width > nodeWidth) nodeWidth = width;
			OrderedTreeNode child = node.firstChild;
			while (child != null)
				{
				nodeWidth = printTree(child, indentation + "    ", fm, nodeWidth);
				child = child.rightSibling;
				}
			}
		return nodeWidth;
		}

	//////////////////////////////////////////////////////////////////////
	private void initializeNodePositions()
		{
		/*
		Walk this tree in levelorder
		and set each node's leftCousin and its tentative position.

		When this returns, walk contains the nodes in level order.
		*/

		//null out the walk vector and return if this tree is empty
		walk = new Vector();
		if (root == null)
			return;

		//do a levelorder walk of this tree, and save it in walk;
		//also set each node's tentative position
		root.y = 0;
		root.x = 0;
		walk.addElement(root);
		OrderedTreeNode child = root.firstChild;
		while (child != null)
			{
			walk.addElement(child);
			child = child.rightSibling;
			}

		OrderedTreeNode lastNode = root;
		for (int i = 1; i < walk.size(); i++)
			{
			//get the next node
			OrderedTreeNode node = (OrderedTreeNode) walk.elementAt(i);

			//set its tentative position relative to its parent
			OrderedTreeNode parent = node.parent;
			node.y = parent.y + 1;
			child = parent.firstChild;
			int rank = -parent.numChildren/2;
			while ((child != null) && (node != child))
				{
				child = child.rightSibling;
				rank++;
				}
			node.x = parent.x + rank;

			//set its leftCousin to the last node examined in levelorder
			//if it's not a sibling
			if ((node.y == lastNode.y) && (node != lastNode.rightSibling))
				node.leftCousin = lastNode;

			//queue its children for later processing
			child = node.firstChild;
			while (child != null)
				{
				walk.addElement(child);
				child = child.rightSibling;
				}

			//save node for the next go round
			lastNode = node;
			}

		}

	//////////////////////////////////////////////////////////////////////
	private float adjustNodePositions(OrderedTreeNode node, float x, String s)
		{
		/*
		Adjust node's horizontal position.

		This is a postorder walk, so all nodes to node's left
		are already in their final position.
		*/

		if (node == null)
			return x;

		//make sure node is to the left of its left cousin or left sibling
		if ((node.leftCousin != null) &&	//it's the leftmost sibling
			(x < (node.leftCousin.x + 1)))	//it's too close
			x = node.leftCousin.x + 1;
		if ((node.leftSibling != null) &&	//it's not leftmost on its level
			(x < (node.leftSibling.x + 1)))	//it's too close
			x = node.leftSibling.x + 1;

		//recursively adjust node's subtrees, and compute their spread
		OrderedTreeNode child = node.firstChild;
		float position = x;
		float minPosition = position;
		float maxPosition = position;
		while (child != null)
			{
			position = adjustNodePositions(child, child.x, s + " ");
			if (minPosition > position) minPosition = position;
			if (maxPosition < position) maxPosition = position;
			child = child.rightSibling;
			}

		//make sure node is midway between its children
		if (node.numChildren > 1)
			node.x = (minPosition + maxPosition)/2;
		else
			node.x = maxPosition;

		return node.x;
		}

	//////////////////////////////////////////////////////////////////////
	private void reset()
		{
		/*
		Reset each node's horizontal position to put the leftmost node at 0

		This assumes that walk contains all this tree's nodes.
		*/

		if (root != null)
			{
			float minX = root.x;
			for (int i = 0; i < numNodes; i++)
				{
				float x = ((OrderedTreeNode) walk.elementAt(i)).x;
				if (x < minX) minX = x;
				}

			for (int i = 0; i < numNodes; i++)
				((OrderedTreeNode) walk.elementAt(i)).x += -minX;
			}
		}

	//////////////////////////////////////////////////////////////////////
	private Point findDisplayedSize()
		{
		/*
		Return the displayed size of this tree.

		This assumes that the tree's nodes are stored in walk.
		*/

		Point max = new Point(0, 0);

		if (root != null)
			for (int i = 0; i < walk.size(); i++)
				{
				OrderedTreeNode node = (OrderedTreeNode) walk.elementAt(i);
				int x = (int) Math.ceil(node.x);
				int y = (int) Math.ceil(node.y);
				if (x > max.x) max.x = x;
				if (y > max.y) max.y = y;
				}

		return max;
		}

	//////////////////////////////////////////////////////////////////////
	private void displayTree(Graphics g, OrderedTreeNode node,
							int nodeWidth, int nodeHeight, int labelOffset,
							FontMetrics fm)
		{
		/*
		Recursively draw node and its subtrees in left-to-right order.
		*/

		if (node == null)
			return;

		//convert node's coordinates to reasonable integers
		int x = (int) (nodeWidth*node.x);
		int y = (int) (2*nodeHeight*node.y);

		//draw it
		//g.drawOval(x, y, nodeWidth, nodeHeight);

		//center the key
		int width = fm.stringWidth(node.key.toString());
		int left = (nodeWidth - width)/2;
		g.drawString((node.key).toString(), x + left, y + labelOffset);

		//connect it to its children, if any, and recursively draw them
		OrderedTreeNode child = node.firstChild;
		while (child != null)
			{
			int childX = (int) (nodeWidth*child.x);
			int childY = (int) (2*nodeHeight*child.y);
			g.drawLine(x + (int) (nodeWidth/2), y + nodeHeight,
						childX + (int) (nodeWidth/2), childY);
			displayTree(g, child, nodeWidth, nodeHeight, labelOffset, fm);
			child = child.rightSibling;
			}
		}

	}

